#define ParentRegistryKey "Software\Martin Prikryl"
#define RegistryKey ParentRegistryKey+"\WinSCP 2"
#define Year 2007

#ifexist "interm\farwinscpsetup.inc.iss"
  #include "interm\farwinscpsetup.inc.iss"
#else
  #define Status "unofficial"
  #define SourceDir ".."
  #define OutputDir ""
#endif

#define MainFileSource SourceDir+"\winscp.dll"
#define EngFileSource SourceDir+"\winscp.lng"
#define PluginSubDirName "WinSCP"

#define Major
#define Minor
#define Rev
#define Build
#expr ParseVersion(MainFileSource, Major, Minor, Rev, Build)
#define Version Str(Major)+"."+Str(Minor)+(Rev > 0 ? "."+Str(Rev) : "")+(Status != "" ? " "+Status : "")

[Setup]
AppId=winscpfar
AppMutex=WinSCPFar
AppName=WinSCP plugin for FAR
AppPublisher=Martin Prikryl
AppPublisherURL=http://winscp.net/eng/docs/far
AppSupportURL=http://winscp.net/forum/
AppUpdatesURL=http://winscp.net/eng/download.php
VersionInfoCompany=Martin Prikryl
VersionInfoDescription=Setup for WinSCP plugin for FAR {#Version}
VersionInfoVersion={#Major}.{#Minor}.{#Rev}.{#Build}
VersionInfoTextVersion={#Version}
VersionInfoCopyright=(c) 2000-{#Year} Martin Prikryl
DefaultDirName={pf}\FAR\Plugins\{#PluginSubDirName}
DisableProgramGroupPage=true
LicenseFile=licence.setup
UninstallDisplayIcon={app}\winscp.ico
OutputDir={#OutputDir}
DisableStartupPrompt=yes
AppVersion={#Version}
AppVerName=WinSCP plugin for FAR {#Version}
OutputBaseFilename=winscpfar{#Major}{#Minor}{#Rev}setup
SolidCompression=yes
PrivilegesRequired=none

[Types]
Name: full; Description: "Full installation"
Name: compact; Description: "Compact installation"
Name: custom; Description: "Custom installation"; Flags: iscustom

[Components]
Name: main; Description: "WinSCP plugin for FAR"; Types: full custom compact
Name: pageant; Description: "Pageant (SSH authentication agent)"; Types: full
Name: puttygen; Description: "PuTTYgen (key generator)"; Types: full

[Files]
Source: "{#MainFileSource}"; DestName: "WinSCP.dll"; DestDir: "{app}"; Components: main; Flags: ignoreversion
Source: "{#EngFileSource}"; DestName: "WinSCP.lng"; DestDir: "{app}"; Components: main; Flags: ignoreversion
Source: "WinSCP.ico"; DestDir: "{app}"; Components: main; Flags: ignoreversion
Source: "licence"; DestName: "licence"; DestDir: "{app}"; Components: main; Flags: ignoreversion
Source: "C:\Program Files\PuTTY\LICENCE"; DestDir: "{app}\PuTTY"; Components: pageant puttygen; Flags: ignoreversion
Source: "C:\Program Files\PuTTY\putty.hlp"; DestDir: "{app}\PuTTY"; Components: pageant puttygen; Flags: ignoreversion
Source: "C:\Program Files\PuTTY\pageant.exe"; DestDir: "{app}\PuTTY"; Components: pageant; Flags: ignoreversion
Source: "C:\Program Files\PuTTY\puttygen.exe"; DestDir: "{app}\PuTTY"; Components: puttygen; Flags: ignoreversion

[InstallDelete]
; Russian language file is no longer maintained
Type: files; Name: "{app}\Russian.lng"

[Code]

procedure ComponentsListClickCheck(Sender: TObject);
begin
  if not WizardForm.ComponentsList.Checked[0] then
    WizardForm.ComponentsList.Checked[0] := True;
end;

procedure InitializeWizard();
var
  S: string;
begin
  // allow installation without requiring user to accept licence
  WizardForm.LicenseAcceptedRadio.Checked := True;
  WizardForm.LicenseAcceptedRadio.Visible := False;
  WizardForm.LicenseLabel1.Visible := False;
  WizardForm.LicenseNotAcceptedRadio.Visible := False;
  WizardForm.LicenseMemo.Top := WizardForm.LicenseLabel1.Top;
  WizardForm.LicenseMemo.Height :=
    WizardForm.LicenseNotAcceptedRadio.Top +
    WizardForm.LicenseNotAcceptedRadio.Height -
    WizardForm.LicenseMemo.Top - 5;

  // prevent the main component to be unchecked
  WizardForm.ComponentsList.OnClickCheck := @ComponentsListClickCheck;

  if RegQueryStringValue(HKCU, 'Software\Far\PluginsCache\Plugin0', 'Name', S) then
  begin
    S := ExtractFileDir(ExtractFileDir(S));
    if CompareText(ExtractFileName(S), 'Plugins') = 0 then
      WizardForm.DirEdit.Text := AddBackslash(S) + '{#PluginSubDirName}';
  end;
end;
